import platform
from setuptools import Extension

import numpy
from Cython.Build import cythonize

compile_args = []
link_args = []

pf = platform.system()
if pf == "Windows":
    # for MSVC
    compile_args = ["/std:c++14", "/DNOMINMAX", "/O2", "/openmp"]
elif pf == "Darwin":
    # for clang
    compile_args = ["-std=c++14", "-O2", "-march=native", "-Xpreprocessor", "-fopenmp"]
    link_args = ["-lomp"]
elif pf == "Linux":
    # for gcc
    compile_args = ["-std=c++14", "-Ofast", "-march=native", "-fopenmp"]
    link_args = ["-fopenmp"]

ext_modules = [
    Extension(
        name="ubo2014_cy",
        sources=["btf_extractor/ubo2014.pyx"],
        include_dirs=[numpy.get_include(), "btf_extractor/c_ext"],
        define_macros=[("BTF_IMPLEMENTATION", "1"), ("NPY_NO_DEPRECATED_API", "1")],
        extra_compile_args=compile_args,
        extra_link_args=link_args,
        language="c++",
    )
]


def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """
    setup_kwargs.update(
        {"ext_modules": cythonize(ext_modules)}
    )
    return setup_kwargs


if __name__ == "__main__":
    build({})



# cython: language_level=3
from itertools import product
from typing import Any, Tuple

cimport cython
import numpy as np
cimport numpy as np
from cython.parallel import prange

from btf_extractor.c_ext.btf cimport BTF, Spectrum, Vector3
from btf_extractor.c_ext.btf cimport BTFFetchSpectrum, DestroyBTF, LoadBTF

DTYPE_F32 = np.float32
ctypedef np.float32_t DTYPE_F32_t


cdef class Ubo2014:
    """BTFDBBのbtfファイルから角度や画像を取り出す

    角度は全て度数法(degree)を用いている。
    画像の実体はopencvと互換性のあるndarray形式(BGR, channels-last)で出力する。

    `col_decode_range`、`row_decode_range`は負の値による指定も可能。
    `img_shape`の範囲外を指定しようとすると`ValueError`を返す。
    `*_decode_range`の書き換えで`img_shape`は変更されない。

    Attributes:
        btf_filepath (str):
            コンストラクタに指定したbtfファイルパス。読み取り専用。
        img_shape (tuple[float,float,float]):
            btfファイルに含まれている画像のshape。読み取り専用。
        angles_set (set[tuple[float,float,float,float]]):
            btfファイルに含まれる画像の角度条件の集合。読み取り専用。
        col_decode_range (tuple[int,int]):
            `angles_to_image()`でデコードする列の範囲。読み書き可能。
        row_decode_range (tuple[int,int]):
            `angles_to_image()`でデコードする行の範囲。読み書き可能。

    Example:
        >>> btf = Ubo2014("carpet01_resampled_W400xH400_L151xV151.btf")
        >>> print(btf.img_shape)
        (400, 400, 3)
        >>> angles_list = list(btf.angles_set)
        >>> print(angles_list[0])
        (60.0, 270.0, 60.0, 135.0)
        >>> image = btf.angles_to_image(*angles_list[0])
        >>> print(image.shape)
        (400, 400, 3)
        >>> print(image.dtype)
        float32
        >>> btf.row_decode_range = (136, 264)
        >>> btf.col_decode_range = (136, -136)
        >>> print(btf.angles_to_image(*angles_list[0]).shape)
        (128, 128, 3)
        >>> print(btf.img_shape)
        (400, 400, 3)
        >>> btf.row_decode_range = (136, 401)
        ValueError: out of range
        >>> btf.row_decode_range = ("136", "264")
        TypeError: indices must be integers
        >>> btf.row_decode_range = (136, 264, 1)
        ValueError: too many values to unpack (expected 2)
    """
    cdef readonly str btf_filepath
    cdef BTF* __raw_btf
    cdef readonly tuple img_shape
    cdef readonly frozenset angles_set
    cdef readonly dict angles_vs_index_dict
    cdef Py_ssize_t row_decode_begin
    cdef Py_ssize_t row_decode_length
    cdef Py_ssize_t col_decode_begin
    cdef Py_ssize_t col_decode_length

    def __cinit__(self, btf_filepath: str) -> None:
        """使用するbtfファイルを指定する"""
        if not type(btf_filepath) is str:
            raise TypeError("`btf_filepath` is not str")
        self.btf_filepath = btf_filepath

        # C++のstringの変換にはポインタの参照が必要。
        # なので、`LoadBTF(btf_filepath.encode(), NULL)`はコンパイルエラーになる。
        cdef bytes py_byte_string = btf_filepath.encode()
        self.__raw_btf = LoadBTF(py_byte_string, NULL)

        # self.__raw_btf.ChannelCountを3と仮定する。
        self.img_shape = (self.__raw_btf.Width, self.__raw_btf.Height, 3)
        self.row_decode_begin = 0
        self.row_decode_length = self.__raw_btf.Width
        self.col_decode_begin = 0
        self.col_decode_length = self.__raw_btf.Height
        cdef Py_ssize_t num_views = self.__raw_btf.ViewCount
        cdef Py_ssize_t num_lights = self.__raw_btf.LightCount

        # *_in_cartesian: 直交座標系で格納された角度情報
        cdef np.ndarray[DTYPE_F32_t, ndim=2] _lights_in_cartesian = np.empty(
            (num_lights, 3), dtype=DTYPE_F32
        )
        cdef Vector3 light
        for light_idx in range(num_lights):
            light = self.__raw_btf.Lights[light_idx]
            _lights_in_cartesian[light_idx,0] = light.x
            _lights_in_cartesian[light_idx,1] = light.y
            _lights_in_cartesian[light_idx,2] = light.z

        cdef np.ndarray[DTYPE_F32_t, ndim=2] _views_in_cartesian = np.empty(
            (num_views, 3), dtype=DTYPE_F32
        )
        cdef Vector3 view
        for view_idx in range(num_views):
            view = self.__raw_btf.Views[view_idx]
            _views_in_cartesian[view_idx,0] = view.x
            _views_in_cartesian[view_idx,1] = view.y
            _views_in_cartesian[view_idx,2] = view.z

        # *_in_spherical: 球面座標系で格納された角度情報
        cdef np.ndarray[DTYPE_F32_t, ndim=2] lights_in_spherical = (
            self._cartesian_to_spherical(_lights_in_cartesian)
        )

        cdef np.ndarray[DTYPE_F32_t, ndim=2] views_in_spherical = (
            self._cartesian_to_spherical(_views_in_cartesian)
        )

        # cdef list angles_list
        cdef list angles_list = [
            (float(l[0]), float(l[1]), float(v[0]), float(v[1]))
            for l, v in product(lights_in_spherical, views_in_spherical)
        ]

        # このBTFファイルに含まれる(tl, pl, tv, pv)の組からindexを引く辞書
        self.angles_vs_index_dict = {
            angles: (lidx, vidx)
            for angles, (lidx, vidx)
            in zip(angles_list, product(range(num_lights), range(num_views)))
        }

        # このBTFファイルに含まれる(tl, pl, tv, pv)の組が格納されたset
        self.angles_set = frozenset(angles_list)

    def __dealloc__(self) -> None:
        DestroyBTF(self.__raw_btf)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef np.ndarray[DTYPE_F32_t, ndim=2] _cartesian_to_spherical(
        self, np.ndarray[DTYPE_F32_t, ndim=2] xyz
    ):
        """直交座標系から球面座標系に変換する"""
        cdef np.ndarray[DTYPE_F32_t, ndim=1] theta
        cdef np.ndarray[DTYPE_F32_t, ndim=1] phi
        theta = np.rad2deg(np.arccos(xyz[:,2]))
        # -180から+180を、0から360に収める
        phi = np.rad2deg(np.arctan2(xyz[:,1], xyz[:,0])) % 360
        # UBO2014の角度情報の精度は0.5度単位なので小数点以下1桁に丸める
        return np.array((theta, phi)).T.round(1)

    @property
    def row_decode_range(self) -> Tuple[int, int]:
        return (self.row_decode_begin, self.row_decode_end)

    @row_decode_range.setter
    def row_decode_range(self, row_index: Tuple[int, int]) -> None:
        if not (type(row_index[0]) is int and type(row_index[1]) is int):
            raise TypeError("indices must be integers")

        cdef int left, right
        left, right = row_index
        if left < 0:
            left = self.__raw_btf.Width + left
        if right < 0:
            right = self.__raw_btf.Width + right

        if right < self.__raw_btf.Width and left < right:
            self.row_decode_begin = left
            self.row_decode_length = right - left
        else:
            raise ValueError("out of range")

    @property
    def col_decode_range(self) -> Tuple[int, int]:
        return (self.col_decode_begin, self.col_decode_end)

    @col_decode_range.setter
    def col_decode_range(self, col_index: Tuple[int, int]) -> None:
        if not (type(col_index[0]) is int and type(col_index[1]) is int):
            raise TypeError("indices must be integers")

        cdef int top, bottom
        top, bottom = col_index
        if top < 0:
            top = self.__raw_btf.Height + top
        if bottom < 0:
            bottom = self.__raw_btf.Height + bottom

        if bottom < self.__raw_btf.Height and top < bottom:
            self.col_decode_begin = top
            self.col_decode_length = bottom - top
        else:
            raise ValueError("out of range")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef np.ndarray[DTYPE_F32_t, ndim=3] _index_to_image_skip_validation(
        self, int light_idx, int view_idx
    ):
        cdef Py_ssize_t col_begin = self.col_decode_begin
        cdef Py_ssize_t row_begin = self.row_decode_begin
        cdef Py_ssize_t height = self.col_decode_length
        cdef Py_ssize_t width = self.row_decode_length
        cdef Py_ssize_t lidx = light_idx
        cdef Py_ssize_t vidx = view_idx
        cdef Spectrum spec
        cdef np.ndarray[DTYPE_F32_t, ndim=3] img3d = np.empty(
            (height, width, 3), dtype=DTYPE_F32
        )
        cdef float[:, :, ::1] img3d_view = img3d
        cdef int x, y
        for y in prange(height, nogil=True, schedule="dynamic"):
            for x in range(width):
                spec = BTFFetchSpectrum(
                    self.__raw_btf, lidx, vidx, x+row_begin, y+col_begin
                )
                img3d_view[y,x,0] = spec.z
                img3d_view[y,x,1] = spec.y
                img3d_view[y,x,2] = spec.x
        return img3d

    def angles_to_image(
        self, tl: float, pl: float, tv: float, pv: float
    ) -> np.ndarray:
        """`tl`, `pl`, `tv`, `pv`の角度条件の画像をndarray形式で返す"""
        cdef tuple key = (tl, pl, tv, pv)
        cdef tuple indices = self.angles_vs_index_dict.get(key)
        if not indices:
            raise ValueError(
                f"Condition {key} does not exist in '{self.btf_filepath}'."
            )
        return self._index_to_image_skip_validation(indices[0], indices[1])

    @cython.nonecheck(False)
    def _angleindex_to_image_unsafe(
        self, light_idx: int, view_idx: int
    ) -> np.ndarray:
        """`light_idx`, `view_idx`で指定される角度条件の画像をndarray形式で返す

        値(引数)の妥当性をチェックしない。不正な値を入力するとクラッシュする。
        """
        return self._index_to_image_skip_validation(light_idx, view_idx)



ctypedef unsigned int uint32_t

cdef extern from "btf.hh" nogil:
    struct Vector3:
        float x, y, z

    ctypedef Vector3 Spectrum

    struct BTF:
        uint32_t ChannelCount
        uint32_t Width, Height

        Vector3 *Lights, *Views
        uint32_t LightCount, ViewCount

    struct BTFExtra:
        pass

    cdef Spectrum BTFFetchSpectrum(
        const BTF *btf,
        uint32_t light_vert, uint32_t view_vert,
        uint32_t x, uint32_t y
    )
    cdef BTF *LoadBTF(const char *file_path, BTFExtra *extra)
    cdef void DestroyBTF(BTF *btf)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['btf_extractor', 'btf_extractor.c_ext']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29,<0.30',
 'imageio>=2.6,<3.0',
 'nptyping>=1,<2',
 'numpy>=1.19,<2.0',
 'simplejpeg>=1.3']

setup_kwargs = {
    'name': 'btf-extractor',
    'version': '1.7.0rc1',
    'description': 'Extract UBO BTF archive format(UBO2003, ATRIUM, UBO2014).',
    'long_description': '# BTF Extractor\n[![PyPI version](https://img.shields.io/pypi/v/btf-extractor?style=flat-square)](https://pypi.org/project/btf-extractor/#history)\n[![GitHub version](https://img.shields.io/github/v/tag/2-propanol/BTF_extractor?style=flat-square)](https://github.com/2-propanol/BTF_extractor/releases)\n[![Python Versions](https://img.shields.io/pypi/pyversions/btf-extractor?style=flat-square)](https://pypi.org/project/btf-extractor/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-black?style=flat-square)](https://github.com/psf/black)\n\nExtract UBO BTF archive format([UBO2003](https://cg.cs.uni-bonn.de/en/projects/btfdbb/download/ubo2003/), [ATRIUM](https://cg.cs.uni-bonn.de/en/projects/btfdbb/download/atrium/), [UBO2014](https://cg.cs.uni-bonn.de/en/projects/btfdbb/download/ubo2014/)).\n\nThis repository uses [zeroeffects/btf](https://github.com/zeroeffects/btf)\'s [btf.hh](https://github.com/zeroeffects/btf/blob/master/btf.hh) (MIT License).\n\nExtract to ndarray compatible with openCV(BGR, channels-last).\n\n## Install\n```bash\npip install btf-extractor\n```\n\nThis package uses the [Cython](https://cython.readthedocs.io/en/latest/src/quickstart/install.html).\nTo install this package, a C++ and OpenMP build environment is required.\n\n### Build is tested on\n- Windows 10 20H2 + MSVC v14.28 ([Build Tools for Visual Studio 2019](https://visualstudio.microsoft.com/downloads/))\n- MacOS 11(Big Sur) + clang 13.0.0 (`xcode-select --install`) + libomp (`brew install libomp`)\n- Ubuntu 20.04 + GCC 9.3.0 ([build-essential](https://packages.ubuntu.com/focal/build-essential))\n\n## Example\n```python\n>>> from btf_extractor import Ubo2003, AtriumHdr, Ubo2014\n\n>>> btf = Ubo2003("UBO_CORDUROY256.zip")\n>>> angles_list = list(btf.angles_set)\n>>> print(angles_list[0])\n(0, 0, 0, 0)\n>>> image = btf.angles_to_image(*angles_list[0])\n>>> print(image.shape)\n(256, 256, 3)\n>>> print(image.dtype)\nuint8\n\n>>> btf = AtriumHdr("CEILING_HDR.zip")\n>>> angles_list = list(btf.angles_set)\n>>> print(angles_list[0])\n(0, 0, 0, 0)\n>>> image = btf.angles_to_image(*angles_list[0])\n>>> print(image.shape)\n(256, 256, 3)\n>>> print(image.dtype)\nfloat32\n\n>>> btf = Ubo2014("carpet01_resampled_W400xH400_L151xV151.btf")\n>>> print(btf.img_shape)\n(400, 400, 3)\n>>> angles_list = list(btf.angles_set)\n>>> print(angles_list[0])\n(60.0, 270.0, 60.0, 135.0)\n>>> image = btf.angles_to_image(*angles_list[0])\n>>> print(image.shape)\n(400, 400, 3)\n>>> print(image.dtype)\nfloat32\n```\n\n## Supported Datasets\n### UBO2003\n6561 images, 256x256 resolution, 81 view and 81 light directions.\n\n![ubo2003](https://user-images.githubusercontent.com/42978570/114306638-59518580-9b17-11eb-9961-baa775ab235f.jpg)\n> Mirko Sattler, Ralf Sarlette and Reinhard Klein "[Efficient and Realistic Visualization of Cloth](http://cg.cs.uni-bonn.de/de/publikationen/paper-details/sattler-2003-efficient/)", EGSR 2003.\n\n### ATRIUM (non-HDR and HDR)\n6561 images, 800x800 resolution, 81 view and 81 light directions.\n\n![atrium](https://user-images.githubusercontent.com/42978570/114306641-5c4c7600-9b17-11eb-8251-9a4a92a16b55.jpg)\n\n### UBO2014\n22,801 images, 512x512(400x400) resolution, 151 view and 151 light directions.\n\n![ubo2014](https://user-images.githubusercontent.com/42978570/114306647-5f476680-9b17-11eb-9fb6-5332e104f341.jpg)\n> [Michael Weinmann](https://cg.cs.uni-bonn.de/en/people/dr-michael-weinmann/), [Juergen Gall](http://www.iai.uni-bonn.de/~gall/) and [Reinhard Klein](https://cg.cs.uni-bonn.de/en/people/prof-dr-reinhard-klein/). "[Material Classification based on Training Data Synthesized Using a BTF Database](https://cg.cs.uni-bonn.de/de/publikationen/paper-details/weinmann-2014-materialclassification/)", accepted at ECCV 2014.\n',
    'author': '2-propanol',
    'author_email': 'nuclear.fusion.247@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/2-propanol/btf_extractor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



"""BTFDBBのzipファイルをzipファイルのまま使用するためのライブラリ

BTFDBB UBO2003(*)形式, ATRIUM(**)形式のzipファイルを参照し、
・zipファイルに含まれる角度情報の取得
・「撮影条件の角度(tl, pl, tv, pv)」から
　「画像の実体(ndarray形式(BGR, channels-last))」を取得
する関数を提供する

(*) http://cg.cs.uni-bonn.de/en/projects/btfdbb/download/ubo2003/
(**) http://cg.cs.uni-bonn.de/en/projects/btfdbb/download/atrium/
"""
from collections import Counter
from sys import stderr
from typing import Any, Tuple
from zipfile import ZipFile

import imageio
import numpy as np
from nptyping import NDArray
from simplejpeg import decode_jpeg

AnglesTuple = Tuple[int, int, int, int]
BGRImage = NDArray[(Any, Any, 3), np.uint8]
BGRImageHDR = NDArray[(Any, Any, 3), np.float32]


class Ubo2003:
    """BTFDBBのzipファイルから角度や画像を取り出す

    角度は全て度数法(degree)を用いている。
    zipファイルに含まれる角度情報の順番は保証せず、並べ替えもしない。
    `angles_set`には`list`ではなく、順序の無い`set`を用いている。

    画像の実体はopencvと互換性のあるndarray形式(BGR, channels-last)で出力する。

    zipファイル要件:
        f"tl{tl:03} pl{pl:03} tv{tv:03} pv{pv:03}.jpg"を格納している。

    Attributes:
        zip_filepath (str): コンストラクタに指定したzipファイルパス。
        angles_set (set[tuple[int,int,int,int]]): zipファイルに含まれる画像の角度条件の集合。

    Example:
        >>> btf = Ubo2003("UBO_CORDUROY256.zip")
        >>> angles_list = list(btf.angles_set)
        >>> print(angles_list[0])
        (0, 0, 0, 0)
        >>> image = btf.angles_to_image(*angles_list[0])
        >>> print(image.shape)
        (256, 256, 3)
        >>> print(image.dtype)
        uint8
    """

    def __init__(self, zip_filepath: str) -> None:
        """使用するzipファイルを指定する

        指定したzipファイルに角度条件の重複がある場合、
        何が重複しているか表示し、`RuntimeError`を投げる。
        """
        self.zip_filepath = zip_filepath
        self.__z = ZipFile(zip_filepath)

        # ファイルパスは重複しないので`filepath_set`はsetで良い
        filepath_set = {path for path in self.__z.namelist() if path.endswith(".jpg")}
        self.__angles_vs_filepath_dict = {
            self._filename_to_angles(path): path for path in filepath_set
        }
        self.angles_set = frozenset(self.__angles_vs_filepath_dict.keys())

        # 角度条件の重複がある場合、何が重複しているか調べる
        if len(filepath_set) != len(self.angles_set):
            angles_list = [self._filename_to_angles(path) for path in filepath_set]
            angle_collection = Counter(angles_list)
            for angles, counter in angle_collection.items():
                if counter > 1:
                    print(
                        f"[BTF-Extractor] '{self.zip_filepath}' has"
                        + f"{counter} files with condition {angles}.",
                        file=stderr,
                    )
            raise RuntimeError(f"'{self.zip_filepath}' has duplicated conditions.")

    @staticmethod
    def _filename_to_angles(filename: str) -> AnglesTuple:
        """ファイル名(orパス)から角度(`int`)のタプル(`tl`, `pl`, `tv`, `pv`)を取得する"""
        # ファイルパスの長さの影響を受けないように後ろから数えている
        tl = int(filename[-25:-22])
        pl = int(filename[-19:-16])
        tv = int(filename[-13:-10])
        pv = int(filename[-7:-4])
        return (tl, pl, tv, pv)

    def angles_to_image(self, tl: int, pl: int, tv: int, pv: int) -> BGRImage:
        """`tl`, `pl`, `tv`, `pv`の角度条件の画像をndarray形式で返す

        `filename`が含まれるファイルが存在しない場合は`ValueError`を投げる。
        """
        key = (tl, pl, tv, pv)
        filepath = self.__angles_vs_filepath_dict.get(key)
        if not filepath:
            raise ValueError(
                f"Condition {key} does not exist in '{self.zip_filepath}'."
            )

        with self.__z.open(filepath) as f:
            return decode_jpeg(f.read(), colorspace="BGR")


class AtriumHdr:
    """ATRIUM(HDR)のzipファイルから角度や画像を取り出す

    角度は全て度数法(degree)を用いている。
    zipファイルに含まれる角度情報の順番は保証せず、並べ替えもしない。
    `angles_set`には`list`ではなく、順序の無い`set`を用いている。

    画像の実体はopencvと互換性のあるndarray形式(BGR, channels-last)で出力する。

    zipファイル要件:
        f"tl{tl:03} pl{pl:03} tv{tv:03} pv{pv:03}.hdr"を格納している。

    Attributes:
        zip_filepath (str): コンストラクタに指定したzipファイルパス。
        angles_set (set[tuple[int,int,int,int]]): zipファイルに含まれる画像の角度条件の集合。

    Example:
        >>> btf = AtriumHdr("CEILING_HDR.zip")
        >>> angles_list = list(btf.angles_set)
        >>> print(angles_list[0])
        (0, 0, 0, 0)
        >>> image = btf.angles_to_image(*angles_list[0])
        >>> print(image.shape)
        (256, 256, 3)
        >>> print(image.dtype)
        float32
    """

    FIXED_HEADER = "-Y 256 +X 256".encode()

    def __init__(self, zip_filepath: str) -> None:
        """使用するzipファイルを指定する

        指定したzipファイルに角度条件の重複がある場合、
        何が重複しているか表示し、`RuntimeError`を投げる。
        """
        self.zip_filepath = zip_filepath
        self.__z = ZipFile(zip_filepath)

        # ファイルパスは重複しないので`filepath_set`はsetで良い
        filepath_set = {path for path in self.__z.namelist() if path.endswith(".hdr")}
        self.__angles_vs_filepath_dict = {
            self._filename_to_angles(path): path for path in filepath_set
        }
        self.angles_set = frozenset(self.__angles_vs_filepath_dict.keys())

        # 角度条件の重複がある場合、何が重複しているか調べる
        if len(filepath_set) != len(self.angles_set):
            angles_list = [self._filename_to_angles(path) for path in filepath_set]
            angle_collection = Counter(angles_list)
            for angles, counter in angle_collection.items():
                if counter > 1:
                    print(
                        f"[BTF-Extractor] '{self.zip_filepath}' has"
                        + f"{counter} files with condition {angles}.",
                        file=stderr,
                    )
            raise RuntimeError(f"'{self.zip_filepath}' has duplicated conditions.")

    @staticmethod
    def _filename_to_angles(filename: str) -> AnglesTuple:
        """ファイル名(orパス)から角度(`int`)のタプル(`tl`, `pl`, `tv`, `pv`)を取得する"""
        # ファイルパスの長さの影響を受けないように後ろから数えている
        tl = int(filename[-25:-22])
        pl = int(filename[-19:-16])
        tv = int(filename[-13:-10])
        pv = int(filename[-7:-4])
        return (tl, pl, tv, pv)

    def angles_to_image(self, tl: int, pl: int, tv: int, pv: int) -> BGRImageHDR:
        """`tl`, `pl`, `tv`, `pv`の角度条件の画像をndarray形式で返す

        `filename`が含まれるファイルが存在しない場合は`ValueError`を投げる。
        """
        key = (tl, pl, tv, pv)
        filepath = self.__angles_vs_filepath_dict.get(key)
        if not filepath:
            raise ValueError(
                f"Condition {key} does not exist in '{self.zip_filepath}'."
            )

        with self.__z.open(filepath) as f:
            # "+X 256 -Y 256"を"-Y 256 +X 256"に書き換える
            raw_bytes = f.read()
            fixed_bytes = raw_bytes[:53] + self.FIXED_HEADER + raw_bytes[66:]
            return imageio.imread(fixed_bytes, format="HDR-FI")[::-1]
